# DataAcquisition.py
# supervisory control and data acquisition

#!/usr/bin/python3
import os
import time
import json
import dbus
import uuid
import shutil
import socket
import base64
import hashlib
import logging
import datetime
import threading
import configparser
from datetime import datetime
from urllib import parse, request
from datetime import datetime
from PyQt5.QtCore import QSettings
from binascii import a2b_hex
from Crypto.PublicKey import RSA
from Crypto.Cipher import PKCS1_OAEP
from json.decoder import JSONDecodeError
from dbus.exceptions import DBusException
from SystemUpdater.Core.UpdaterConfigParser import UpgradeConfig
from apscheduler.schedulers.blocking import BlockingScheduler

LOCALTIDDIR  = "/var/lib/kylin-system-updater/"
LOCALTIDFILE = "tidfile.conf"
MSGSNDDIR    = "/var/lib/kylin-system-updater/sendinfos/"
FILEPATH = "/var/lib/kylin-system-updater/sendinfos/Info_ICBC.json"

class UpdateMsgCollector():
    def __init__(self, manager=None):
        self.UploadMessage = {}
        self.PackageInfo = {}
        self.waitSendList = []
        self.updateManager = manager
        # 转换 & 加密
        self.convertor = FormatConvert(self)
        # 发送器
        self.sender = MessageSend(self)

        self.scheduler = BlockingScheduler()
        self.sendinfo_job = self.scheduler.add_job(self.sender._ReadFromFile, 'interval', kwargs={"json_path":FILEPATH}, hours=24, minutes=5, seconds=20)
        self.timer = TimerThread(self.scheduler)

        logging.info("Initialize Update MessageSend Collector to success...")
        # 配置文件
        # self.configs = UpgradeConfig(LOCALTIDDIR, name=LOCALTIDFILE)

    def GenUploadMessage(self, dict_message):
        UploadMessage = {}
        # 获取将要上传的数据,获取东八区时间
        UploadMessage['createTimeStamp'] = get_east_8_time()
        try:
            for key in dict_message.keys():
                if "packageName" != key:
                    UploadMessage[key] = dict_message[key]
            UploadMessage['UUID'] = str(uuid.uuid1())
        except Exception as e:
            logging.error(str(e))

        json_UploadMessage = self.convertor.dictConvertJson(UploadMessage)

        logging.info('Generate UploadMessage: %s.',json_UploadMessage)
        self.UploadMessage = UploadMessage.copy()
        UploadMessage.clear()

    def GenPackageInfo(self, messageType, packageName):
        PackageInfo = {}
        PackageInfo['messageType'] = str(messageType)
        PackageInfo['packageName'] = str(packageName)
        # 获取本地tid
        self.sender.GetLocalTid()
        PackageInfo["tid"] = str(self.sender.localtid)
        
        json_PackageInfo = self.convertor.dictConvertJson(PackageInfo)

        logging.info('Generate PackageInfo: %s.',json_PackageInfo)
        self.PackageInfo = PackageInfo.copy()
        PackageInfo.clear()

    def setUploadMessage(self, KeyValue):
        # FIXME:数据获取顺序问题
        pass

    def UpdateMsg(self, messageType, json_message):
        # para: messageType(消息类型):  "UpdateInfos"、 "InstallInfos"、 "RemoveInfo"
        # para: dict_message(数据内容): 必须包含 "packageName"、"source"", 采集器会进行检测
        dict_message = self.convertor.JsonConvertDict(json_message)
        if messageType == "":
            messageType = "SystemUpdate"
        elif messageType == "UpdateInfos":
            # 获取系统激活码
            Activation_code = GetActivationCode()
            dict_message['RegisterNumber'] = Activation_code
        if type(dict_message) != type(dict) and "packageName" not in dict_message.keys():
            raise AttributeError("'%s' object has no attribute '%s'" % ("dict_message", "packageName"))
            return False
        
        # 生成UploadMessage与PackageInfo
        try:
            self.GenPackageInfo(messageType, dict_message["packageName"])
            self.GenUploadMessage(dict_message)
        except Exception as e:
            logging.error(str(e))
        
        # sha256
        json_UploadMessage = self.convertor.dictConvertJson(self.UploadMessage)
        json_PackageInfo   = self.convertor.dictConvertJson(self.PackageInfo)
        shaValue = self.convertor.Sha256Value(json_UploadMessage) 
        encodeMsg = self.convertor.EncodeRSAtoBase64(shaValue)
        
        # dbus发送
        self.sender.MsgSendToServer(json_UploadMessage, json_PackageInfo, encodeMsg)

        
class FormatConvert():
    def __init__(self, DataCollector):
        #秘钥
        self.publickey = UniqueKey()
        self.collector = DataCollector

    def dictConvertJson(self, dict_msg):
        #字典转换为json格式字符串
        json_file = ''
        try:
            json_file = json.dumps(dict_msg)
        except JSONDecodeError as e:
            logging.error(str(e))
        return json_file
    
    def JsonConvertDict(self, json_file):
        # json格式字符串转换为字典
        dict_file = {}
        try:
            dict_file = json.loads(json_file)
        except JSONDecodeError as e:
            logging.error(str(e))
        return dict_file
    
    def Sha256Value(self, json_file):
        # 计算sha256值
        hsobj = hashlib.sha256()
        try:
            hsobj.update(json_file.encode("utf-8"))
        except ValueError as e:
            logging.error("SHA256 value error: %s.",str(e))
        except Exception as e:
            logging.error(str(e))
        
        return hsobj.hexdigest()

    def EncodeRSAtoBase64(self, value):
        # 将value进行RSA加密并base64转码
        try:
            # 计算hex值
            value_hex = a2b_hex(value)
            # 加载公钥,填充格式OAEP
            uniqueKey = self.publickey.keyvalue.encode('utf-8')
            uniqueKeyorig = base64.b64decode(uniqueKey)  # 公钥文件
            rsa_pubkey = RSA.importKey(uniqueKeyorig)    # RSA公钥
            oaep_pub = PKCS1_OAEP.new(rsa_pubkey)        # OAEP填充
            # 加密数据
            encodemsg = oaep_pub.encrypt(value_hex)
            # 加密数据Base64转码
            enMsg = base64.b64encode(encodemsg)   
        except ValueError:
            logging.error("Value error: %s.", value)
        except TypeError:
            logging.error("RSA key has no private half.")
        return enMsg


class MessageSend():
    FAIL_SEND = "Error: unknown input type"
    SUCCESS_SEND = "log sent"

    def __init__(self, DataCollector=None) -> None:
        # self.convertor = FormatConvert()
        if DataCollector == None:
            self.collector = UpdateMsgCollector()
        else:
            self.collector = DataCollector
        # self._ReadFromFile("/var/lib/kylin-system-updater/sendinfos/testMsg.json")
        # self._TimedTransmission()

    def MsgSendToServer(self, UploadMessage, PackageInfo, encodeMsg):
        daqbus = dbus.SystemBus()
        messgae = {}
        info = {}
        try:
            daqobj = daqbus.get_object('org.log.sys_transmit', '/org/log/sys_transmit')
            daqinterface = dbus.Interface(daqobj, dbus_interface='org.log.transmit')
        except DBusException as e:
            logging.error(str(e))

        dict_UploadMessage = self.collector.convertor.JsonConvertDict(UploadMessage)
        dict_PackageInfo = self.collector.convertor.JsonConvertDict(PackageInfo)

        json_UploadMessage = self.collector.convertor.dictConvertJson(dict_UploadMessage)
        json_PackageInfo = self.collector.convertor.dictConvertJson(dict_PackageInfo)

        messgae.update({"packageName": str(dictgetattr(dict_PackageInfo, "packageName", 'None'))})
        messgae.update({"messageType": str(dictgetattr(dict_PackageInfo, "messageType", 'None'))})
        for k in dict_UploadMessage.keys():
            messgae.update({k: str(dictgetattr(dict_UploadMessage, k, 'unKnown'))})
        json_messgae = self.collector.convertor.dictConvertJson(messgae)

        info.update({"time":str(dictgetattr(dict_UploadMessage, 'createTimeStamp', ' '))})
        info.update({"hostname":str("root")})
        info.update({"ip":str("unknown")})
        info.update({"name":str("kylin-system-updater")})
        info.update({"lv":str("debug")})
        info.update({"message":json_messgae})
        json_info = self.collector.convertor.dictConvertJson(info)

        try:
            retval = daqinterface.log_transmit(json_info)
            logging.info("Send value: %s.", retval)
        except AttributeError:
            logging.error("Call UploadMessage: Attribute Error.")
        except Exception as err:
            logging.error(err)
        self.Send_finally(retval, json_info)
    
    def Send_finally(self, retval, json_info):
        # 根据发送结果进行处理
        result = ''
        PackageInfo = self.collector.convertor.JsonConvertDict(json_info)
        #if str(retval) == str(self.FAIL_SEND): 
            # 上传失败写入本地json
        self.WriteToJson(PackageInfo)

    def GetLocalTid(self):
        # 试图获取本地tid
        try:
            if os.path.exists(LOCALTIDDIR+LOCALTIDFILE):
                with open(LOCALTIDDIR+LOCALTIDFILE, "r") as f:
                    self.localtid = f.readline()
                # print(self.localtid)
            else:
                self.localtid = ""
        except Exception as e:
            logging.error(str(e))

    def SaveTid(self, localtid):
        if len(localtid) == 0:
            return 
        try:
            if os.path.exists(LOCALTIDDIR+LOCALTIDFILE):
                with open(LOCALTIDDIR+LOCALTIDFILE, "w+") as f:
                    f.write(str(localtid))
            else:
                with open(LOCALTIDDIR+LOCALTIDFILE, "w+") as f:
                    f.write(str(localtid))
        except Exception as e:
            logging.error(str(e))

    def WriteToJson(self, PackageInfo):
        #发送失败时，写入本地json中定时发送
        json_file = self.collector.convertor.dictConvertJson(PackageInfo)
        # 保存信息
        try: 
            if not os.path.exists(MSGSNDDIR):
                os.mkdir(MSGSNDDIR)
            # 根据messageType保存信息
            with open(MSGSNDDIR+"Info_ICBC"+".json","a") as f:
                f.write(json_file)
                f.write("\n")
        except Exception as e:
            logging.error(str(e))

    def _TimedTransmission(self, file_path = MSGSNDDIR):
        classify_list = [name for name in os.listdir(file_path) if name.endswith(".json")]
        for f in classify_list:
            # 循环发送每一个文件
            self._ReadFromFile(os.path.join(file_path, f))
    
    def _ReadFromFile(self, json_path):
        # logging.info("Starting send info...")
        self.collector.scheduler.pause()
        new_lines = []
        # 从本地文件读取
        if not os.path.exists(json_path):
            self.collector.scheduler.resume()
            # logging.info("End of this sending...")
            return 
        with open(json_path, "r+") as f:
            lines = f.readlines()

        # file is empty and path is exit -> remove file
        if len(lines) == 0 and os.path.exists(json_path):
            os.remove(json_path)
            self.collector.scheduler.resume()
            # logging.info("End of this sending...")
            return 

        #send installinfo or updateinfo
        for line in lines:
            retval = self._file_send_server(line) 
            if str(retval) == str(self.FAIL_SEND): 
            # 上传失败写入本地json
                new_lines.append(line)
        if os.path.exists(json_path):
            os.remove(json_path)
        if len(new_lines) != 0:
            logging.warning("%s messages failed to be sent, please view the uploaded file...", len(new_lines))
            with open(json_path, "w+") as f:
                for line in lines:
                    f.write(line)
        # logging.info("End of this sending...")
        self.collector.scheduler.resume()

    def _file_send_server(self, json):
        dict_msg = self.collector.convertor.JsonConvertDict(json)
        json_msg = self.collector.convertor.dictConvertJson(dict_msg)
        daqbus = dbus.SystemBus()
        try:
            daqobj = daqbus.get_object('org.log.sys_transmit', '/org/log/sys_transmit')
            daqinterface = dbus.Interface(daqobj, dbus_interface='org.log.transmit')
        except DBusException as e:
            logging.error(str(e))
        try:
            retval = daqinterface.log_transmit(json_msg)
            logging.info("Send result: %s.", retval)
        except AttributeError:
            logging.error("Call UploadMessage: Attribute Error.")
        except Exception as e:
            logging.error(e)
        return retval

class TimerThread(threading.Thread):
    def __init__(self, scheduler):
        threading.Thread.__init__(self)
        self.setName("TimerSendInfo")
        self.scheduler = scheduler
    def run(self):
        self.scheduler.start()
    def stop(self):
        self.scheduler.shutdown()

class UniqueKey():
    keyvalue = "LS0tLS1CRUdJTiBQVUJMSUMgS0VZLS0tLS0KTUlJQklqQU5CZ2txaGtpRzl3MEJBUUVGQUFPQ0FR\
OEFNSUlCQ2dLQ0FRRUFzdW1NTFJEdlFNb0tEQkRJODRqSgpqc1A0Mk55V0pWVEZob2Jra3ZiT05j\
dExYTXVzRmo2TzJUblZYU3Z6VlVLSjRqZkpwT2l2WEphOVB5Z2wzYTRnClBzSU40enNCMEdOY0tr\
R3VsS2RrV2x6S3lWQ2xlTzhiQnN6SjkwbTc3cWF6YWg3a1A0TUl0WTVFczBpSkpiR0oKN1MxcERj\
MlJkNnVFQWJLaXJyRTFlNzlFTEd4am5VN2V5NWkyRDE2WWJoZEQwZ2lNa2RHR3piQXBKTWZWRVJR\
TQo1NXorMFVqdS8zSFJhNFY3b3p2TGRPRE5HUURaeWNJU0l3VHBLbFR3RjBxazdCNjVhTUlJenQ1\
dnhOK1lxYU1GClppZFRLNzcxNjdqNEExZ3F3MG45bjlybWVXUGRWZ3dudnRtVXp4Q1krNk05SXpK\
TDI3eWpRUTV1WGQ3RVdMT3IKbndJREFRQUIKLS0tLS1FTkQgUFVCTElDIEtFWS0tLS0tCg=="

def get_east_8_time():
    import time
    # UTC时间
    utc_time = datetime.utcnow() 
    # 转时间字符串
    utc_time = utc_time.strftime("%Y-%m-%d %H:%M:%S.%f")[:-3]
    time_suffix = utc_time.split(".")[1]
    # 字符串转时间元祖
    utc_time = time.strptime(utc_time, "%Y-%m-%d %H:%M:%S.%f")
    # 时间元祖转时间戳
    utc_time = time.mktime(utc_time)
    # 生成东八区时间时间戳
    now_time = utc_time + 8*60*60
    # 时间戳转时间元祖
    now_time = time.localtime(now_time)
    # 时间元祖转字符串
    now_time = time.strftime("%Y-%m-%d %H:%M:%S",now_time)
    now_time = now_time + "." +time_suffix
    return now_time
    # return 0

def GetActivationCode():
    Code = ''
    try:
        bus = dbus.SystemBus()
        obj = bus.get_object('org.freedesktop.activation', '/org/freedesktop/activation')
        interface = dbus.Interface(obj, dbus_interface='org.freedesktop.activation.interface')
        retval,ret = interface.register_number()
        if ret == 0 and len(retval) != 0:
            Code = str(retval)
        elif len(retval) == 0:  # sh 查询
            sh_retval = os.popen("kylin_gen_register").read().strip()
            if len(sh_retval) != 0:
                Code = str(sh_retval)
    except Exception as e:
        logging.error(str(e))
        return ""
    return Code

def dictgetattr(_dict, _key, default_value):
    if type(_dict) != type(dict()):
        return 
    if _key in _dict.keys():
        return _dict[_key]
    else :
        return default_value

if __name__ == "__name__":
    # 执行定时发送
    ms = MessageSend()
    ms._ReadFromFile("/var/lib/kylin-system-updater/sendinfos/testMsg.json")